{-# LANGUAGE BlockArguments #-}

module Day13 (parsePackets, Packet (..), checkPair, sumIndices, day13, parsePacketList, addDividers, part2, findDividers, dividers) where

import Text.Megaparsec
import Text.Megaparsec.Char
import Data.Void (Void)
import Control.Monad (void)
import Data.List (intercalate, elemIndex, sort)
import Data.Maybe (catMaybes, mapMaybe)

data Packet = List [Packet] | Entry Int

instance Eq Packet where
  (Entry a) == (Entry b)  = a == b
  (List [Entry a]) == (Entry b) = a == b
  (Entry a) == (List [Entry b]) = a == b
  (List []) == (Entry _) = False
  (List []) == (List (_:_)) = False
  (List []) == (List []) = True
  (Entry _) == (List [])  = False
  (List (a:as)) == (List (b:bs)) = a == b && List as == List bs
  a@(Entry _) == b@(List _) = List [a] == b
  a@(List _) == b@(Entry _) = a == List [b]
  (List (_:_)) == (List []) = False

instance Ord Packet where
  compare (Entry x) (Entry y) = x `compare` y
  compare (List []) (List []) = EQ
  compare (List []) (List (_:_)) = LT
  compare (List ((Entry x):xs)) (List ((Entry y):ys))
    | x < y = LT
    | x == y = List xs `compare` List ys
    | otherwise = GT
  compare (List (x:xs)) (List (y:ys)) =
    case x `compare` y of
      LT -> LT
      EQ -> List xs `compare` List ys
      GT -> GT
  compare (List (_:_)) (List []) = GT
  compare (Entry x) (List ys) = List [Entry x] `compare` List ys
  compare (List xs) (Entry y)=  List xs `compare` List [Entry y]


newtype PacketPair = PacketPair (Packet, Packet) deriving (Eq)

instance Show PacketPair where
  show (PacketPair (p1, p2)) = show p1 ++ "\n" ++ show p2 ++ "\n"

instance Show Packet where
  show (Entry n) = show n
  show (List ps) = "["  ++ intercalate "," (map show ps) ++ "]"

type Parser = Parsec Void String

parsePackets :: String -> [PacketPair]
parsePackets input = case parse (packetPairs `sepBy` eol) "" input of
  Left err -> error (show err)
  Right result -> result

parsePacketList :: String -> [Packet]
parsePacketList input = case parse (packetLine `sepBy` eol) "" input of
  Left err -> error (show err)
  Right result -> catMaybes result

packetLine :: Parser (Maybe Packet)
packetLine = Just <$> parsePacket <|> return Nothing

packetPairs :: Parser PacketPair
packetPairs = do
    p1 <- parsePacket
    void eol
    p2 <- parsePacket
    void eol
    return $ PacketPair (p1, p2)

parsePacket :: Parser Packet
parsePacket = do
  void $ char '['
  entries <- (parseNum <|> parsePacket) `sepBy` char ','
  void $ char ']'
  return $ List entries
  where
    parseNum = Entry . read <$> some numberChar

checkPair :: PacketPair -> Bool
checkPair (PacketPair (p1, p2)) = p1 <= p2


sumIndices :: [PacketPair] -> Int
sumIndices pairs = sum $ map fst $ filter snd $ zip [1..] $ map checkPair pairs


dividers :: [Packet]
dividers = [List [List [Entry 2]], List [List [Entry 6]]]

addDividers :: [Packet] -> [Packet]
addDividers = (++)  dividers

-- find the 1-based indices of the dividers
findDividers :: [Packet] -> [Int]
findDividers packets = map (+1) positions
  where
    positions = mapMaybe findIndex dividers
    findIndex x = x `elemIndex` packets

part2 :: [Packet] -> Int
part2 packets = product $ findDividers $ sort $ addDividers packets

day13 :: IO ()
day13 = do
  input <- readFile "ressources/day13-input"
  putStrLn "Day13"
  let pairs = parsePackets input
  putStrLn ("Sum of indices: " ++ show (sumIndices pairs))
  let packets = parsePacketList input
  putStrLn ("Decoder key: " ++ show (part2 packets))
